
import itertools, numpy as np, math
import uuid

from . import VectorOps as vec_ops
from . import SetOps as sets
from .Misc import is_numeric, is_zero

__all__ = [
    "permutation_sign",
    "levi_cevita_maps",
    "levi_cevita_tensor",
    "levi_cevita3",
    "levi_cevita_dot",
    "nca_op_deriv",
    "tensordot_deriv",
    "tensorprod_deriv",
    "scalarprod_deriv",
    "inverse_transformation",
    "optimizing_transformation",
    "matinv_deriv",
    "matdet_deriv",
    "matsqrt_deriv",
    "scalarinv_deriv",
    "scalarpow_deriv",
    "tensor_reexpand",
    "tensorops_deriv",
    "vec_norm_unit_deriv",
    "vec_angle_deriv",
    "vec_cross_deriv",
    "vec_anglecos_deriv",
    "vec_anglesin_deriv",
    "vec_dihed_deriv",
    "vec_plane_angle_deriv"
]

def permutation_sign(perm, check=True):
    # essentially a swap sort on perm
    # https://stackoverflow.com/a/73511014
    parity = 1
    perm = np.asanyarray(perm)
    if check:
        perm = np.argsort(np.argsort(perm))
    else:
        perm = perm.copy()
    for i in range(len(perm)):
        while perm[i] != i: # ensure subblock is sorted
            parity *= -1
            j = perm[i]
            perm[i], perm[j] = perm[j], perm[i]
    return parity
def levi_cevita_maps(k):
    perms = sets.permutation_indices(k, k)
    signs = np.array([permutation_sign(p, check=False) for p in perms])
    return perms, signs
def levi_cevita_tensor(k, sparse=False):
    pos, vals = levi_cevita_maps(k)
    if sparse:
        from .Sparse import SparseArray
        a = SparseArray.from_data(
            (
                pos.T,
                vals
            ),
            shape=(k,)*k
        )
    else:
        a = np.zeros((k,)*k, dtype=int)
        a[tuple(pos.T)] = vals
    return a
# levi_cevita3 = levi_cevita_tensor(3)
levi_cevita3 = np.array([
    [[0, 0, 0],
     [0, 0, 1],
     [0, -1, 0]],

    [[0, 0, -1],
     [0, 0, 0],
     [1, 0, 0]],

    [[0, 1, 0],
     [-1, 0, 0],
     [0, 0, 0]]
])

def sylvester_solve(A, B, C):
    A = np.asanyarray(A)
    B = np.asanyarray(B)
    C = np.asanyarray(C)
    n = A.shape[1]
    m = B.shape[0]
    S = np.kron(B.T, np.eye(n)) + np.kron(np.eye(m), A)
    val = np.linalg.solve(S, C.flatten()) # in case it can be cleverer than I can
    return val.reshape(m,n)

def levi_cevita_dot(k, a, /, axes, shared=None):
    pos, vals = levi_cevita_maps(k)
    return vec_ops.semisparse_tensordot((tuple(pos.T), vals, (k,) * k), a, axes, shared=shared)
    # a = np.asanyarray(a)
    # lv_axes = axes[0]
    # a_axes = axes[1]
    # if util.is_numeric(lv_axes):
    #     lv_axes = [lv_axes]
    # if util.is_numeric(a_axes):
    #     a_axes = [a_axes]
    # for ax in reversed(a_axes):
    #     a = np.moveaxis(a, ax, 0)
    #
    # nax = len(lv_axes)
    # if nax == k:
    #     new = np.dot(vals, a[tuple(pos.T)])
    # else:
    #     new_shape = (k,) * (k-nax) + a.shape[nax:]
    #     new = np.zeros(new_shape)
    #     idx = np.setdiff1d(np.arange(k), lv_axes)
    #     for p, v in zip(pos, vals):
    #         new[p[lv_axes]] += v*a[p[idx]]
    # return new

# levi_cevita3.__name__ = "levi_cevita3"
# levi_cevita3.__doc__ = """
#     The 3D Levi-Cevita tensor.
#     Used to turn cross products into matmuls
#     """



def get_nca_shifts(order, k):
    permute_pos = np.arange(k)
    ncombs = math.comb(order, k)
    shifts = np.broadcast_to(np.arange(order)[np.newaxis], (ncombs, order)).copy()
    for i,pos in enumerate(itertools.combinations(range(order), r=k)):
        shifts[i, pos] = permute_pos
    return shifts
def apply_nca_op(op, order, k, A_expansion, B_expansion, deriv_axis, a, b, contract, shared, identical):
    s = order - k
    if s >= len(A_expansion) or k >= len(B_expansion):
        return 0

    A = A_expansion[s]
    B = B_expansion[k]
    for T in [A, B]:
        if (
            is_zero(T)
            or (T.shape == () and T == 0)
        ):
            return 0
    if shared is None:
        shared = 0
    if contract: # axes disappear, so we just account for the shifts
        axes = [[x+s for x in a], [x+k for x in b]]
    else: # axes appeared, so we need to include those in the product
          # we _forced_ axes to be at the end of the arrays since it was too hard
          # to keep track of them otherwise...
          # actually I guess I could have put the derivative axes at the end...
          # and then the axes would never change...but that has other complications
        axes = [
            [shared + i for i in range(s)] + [x+s for x in a],
            [shared + i for i in range(k)] + [x+k for x in b]
        ]

    if shared == 0:
        base = op(A, B, axes=axes)
    else:
        base = op(A, B, axes=axes, shared=shared)

    # next we need to move all of the derivative axes in the second tensor to the beginning
    # so we can symmetrize
    d = deriv_axis + s
    for i in range(k):
        base = np.moveaxis(base, d+i, shared)
    part = [x for x in reversed(sorted([s,k])) if x > 0]
    if len(part) > 1:
        base = nca_symmetrize(base, part, shared=shared, identical=identical)
    return base
def nca_op_order_deriv(op, order, A_expansion, B_expansion, deriv_axis, a, b, contract, shared, identical):
    full = None
    for k in range(order+1):
        term = apply_nca_op(op, order, k, A_expansion, B_expansion, deriv_axis, a, b, contract, shared, identical)
        if full is None:
            full = term
        else:
            full = full + term
    return full
def nca_op_deriv(op,
                 A_expansion,
                 B_expansion,
                 order,
                 axes,
                 contract,
                 shared=None,
                 identical=False
                 ):
    A_expansion = [np.asanyarray(A) for A in A_expansion]
    B_expansion = [np.asanyarray(B) for B in B_expansion]

    a_ax, b_ax = axes
    if isinstance(a_ax, int): a_ax = [a_ax]
    if isinstance(b_ax, int): b_ax = [b_ax]
    a_ax = [ a if a >= 0 else A_expansion[0].ndim + a for a in a_ax ]
    b_ax = [ b if b >= 0 else B_expansion[0].ndim + b for b in b_ax ]

    if contract: # the derivative axis will always be at nA + 1 - num_contracted - the shared axes
        deriv_axis = A_expansion[0].ndim - len(a_ax)
    else:
        # we require that the outer product be ordered
        # so we now which axes to move around
        a_ax = np.sort(a_ax)
        a_dim = A_expansion[0].ndim
        if np.any(a_ax != np.arange(a_dim - len(a_ax), a_dim)):
            raise ValueError("axes {} must be the final axes of A".format(a_ax))
        b_ax = np.sort(b_ax)
        b_dim = B_expansion[0].ndim
        if np.any(b_ax != np.arange(b_dim - len(b_ax), b_dim)):
            raise ValueError("axes {} must be the final axes of B".format(b_ax))
        deriv_axis = a_dim

    if isinstance(order, int):
        order = list(range(1, order+1))

    # if shared is not None:
    #     deriv_axis = deriv_axis - shared

    derivs = [
        nca_op_order_deriv(op, o, A_expansion, B_expansion, deriv_axis, a_ax, b_ax,
                           contract, shared, identical
                           )
        for o in order
    ]

    return derivs

def _deriv_construct(zero_order_handler, expansion_generator, order):
    if is_numeric(order):
        order = list(range(order+1))
    high_order_expansion = expansion_generator([o for o in order if o > 0])

    # a lot of work to make sure we can do derivatives at a specifically targeted order
    n = 0
    final_expansion = []
    for o in order:
        if o == 0:
            final_expansion.append(
                zero_order_handler()
            )
        else:
            final_expansion.append(high_order_expansion[n])
            n += 1
    return final_expansion

def tensordot_deriv(A_expansion, B_expansion,
                    order,
                    axes=None,
                    shared=None,
                    identical=False
                    ):

    if len(A_expansion) == 0 or len(B_expansion) == 0:
        return [0]*((order+1) if is_numeric(order) else len(order))

    if axes is None: axes = [-1, 0]
    if shared is not None and shared > 0:
        base_op = vec_ops.vec_tensordot
    else:
        base_op = lambda a,b,axes=None,shared=None:np.tensordot(a, b, axes=axes)
    op = lambda a,b,axes=None, shared=None: (
        0
            if is_zero(a) or is_zero(b) else
        base_op(a, b, axes=axes, shared=shared)
    )
    return _deriv_construct(
        lambda : op(A_expansion[0], B_expansion[0], axes=axes, shared=shared),
        lambda ords: nca_op_deriv(op,
                     A_expansion, B_expansion,
                     ords,
                     axes=axes,
                     contract=True,
                     shared=shared,
                     identical=identical
                     ),
        order
    )

def prep_prod_arrays(A_expansion, a_ax):
    _ = []
    for n,a in enumerate(A_expansion):
        a = np.asanyarray(a)
        for x in a_ax:
            x = (a.ndim + x) if x < 0 else (x + n)
            a = np.moveaxis(a, x, -1)
        _.append(a)
    return _
def pre_broadcast_prod(A_expansion, B_expansion, axes):

    a_ax, b_ax = axes
    if isinstance(a_ax, int): a_ax = [a_ax]
    if isinstance(b_ax, int): b_ax = [b_ax]

    A_expansion = prep_prod_arrays(A_expansion, a_ax)
    B_expansion = prep_prod_arrays(B_expansion, b_ax)

    expand_dims = ...

def tensorprod_deriv(
        A_expansion, B_expansion,
        order,
        axes=None,
        identical=False
):
    if len(A_expansion) == 0 or len(B_expansion) == 0:
        return [0]*((order+1) if is_numeric(order) else len(order))
    if axes is None:
        axes = [-1, -1]
    elif isinstance(axes, str) and axes == 'all':
        axes = [np.arange(A_expansion[0].ndim), np.arange(B_expansion[0].ndim)]
    a_ax, b_ax = axes
    if isinstance(a_ax, int): a_ax = [a_ax]
    if isinstance(b_ax, int): b_ax = [b_ax]
    # a_ax, b_ax = pre_broadcast_prod(A_expansion, B_expansion, axes)

    shared = A_expansion[0].ndim - len(a_ax)
    # if is_numeric(A_expansion[0]) or is_numeric(B_expansion[0]):
    #     # scalar product
    #     return scalarprod_deriv(A_expansion, B_expansion, order, identical=identical)

    A_expansion = [np.asanyarray(a) for a in A_expansion]
    B_expansion = [np.asanyarray(b) for b in B_expansion]
    op = lambda left, right, axes=None, shared=None: (
        0
            if is_zero(left) or is_zero(right) else
        vec_ops.vec_outer(left, right, axes=axes, order=0)
    )
    return _deriv_construct(
        lambda: op(A_expansion[0], B_expansion[0], axes=axes, shared=shared),
        lambda ords: nca_op_deriv(op,
                                  A_expansion, B_expansion,
                                  ords,
                                  axes=[a_ax, b_ax],
                                  contract=False,
                                  shared=shared,
                                  identical=identical
                                  ),
        order
    )

def _scalar_prod(a, b):
    if is_numeric(a) or is_numeric(b):
        if is_zero(a) or is_zero(b):
            return 0
        else:
            return a * b
    else:
        return vec_ops.vec_outer(a, b, axes=[list(range(a.ndim)), list(range(b.ndim))], order=0)

def scalarprod_deriv(s_expansion, A_expansion,
                     order,
                     identical=False
                     ):
    s_expansion = [np.asarray(s) for s in s_expansion]
    A_expansion = [np.asarray(a) for a in A_expansion]
    s0_shape = s_expansion[0].shape if len(s_expansion) > 0 else ()
    A0_shape = A_expansion[0].shape if len(A_expansion) > 0 else ()
    shared = min([len(s0_shape), len(A0_shape)])
    return tensorprod_deriv(
        s_expansion,
        A_expansion,
        order=order,
        identical=identical,
        axes=[
            np.arange(shared, s_expansion[0].ndim),
            np.arange(shared, A_expansion[0].ndim)
        ]
    )
    if is_numeric(order):
        order = list(range(order+1))
    suborder = [o for o in order if o > 0]
    # s_expansion, A_expansion = pre_broadcast_scalar_mult(s_expansion, A_expansion)
    if is_numeric(A_expansion[0]):
        if is_numeric(s_expansion[0]):
            base_expansion = scalarprod_deriv(s_expansion, A_expansion[1:],
                                              [o - 1 for o in suborder],
                                              identical=identical)
        else:
            shared = min([s_expansion[0].ndim, A_expansion[0].ndim])
            base_expansion = tensorprod_deriv(s_expansion, A_expansion[1:],
                                              [o - 1 for o in suborder],
                                              identical=identical,
                                              axes=[
                                                  np.arange(shared, s_expansion[0].ndim),
                                                  np.arange(shared, A_expansion[0].ndim)
                                              ])
        rem_expansion = [
            _scalar_prod(s_expansion[o], A_expansion[0])
                if len(s_expansion) > o else 0
            for o in suborder
        ]
    elif not is_numeric(s_expansion[0]):
        shared = min([s_expansion[0].ndim, A_expansion[0].ndim])
        return tensorprod_deriv(s_expansion, A_expansion,
                                order,
                                identical=identical,
                                axes=[
                                    np.arange(shared, s_expansion[0].ndim),
                                    np.arange(shared, A_expansion[0].ndim)
                                ])
    else:
        base_expansion = tensorprod_deriv(s_expansion[1:], A_expansion,
                                          order=[o - 1 for o in suborder],
                                          axes=[
                                              np.arange(s_expansion[1].ndim),
                                              np.arange(A_expansion[0].ndim)
                                          ],
                                          identical=identical)
        rem_expansion = [
            _scalar_prod(s_expansion[0], A_expansion[o]) if len(A_expansion) > o else 0
            for o in suborder
        ]
    high_order_expansion = [x + y for x, y in zip(base_expansion, rem_expansion)]

    # a lot of work to make sure we can do derivatives at a specifically targeted order
    n = 0
    final_expansion = []
    for o in order:
        if o == 0:
            final_expansion.append(_scalar_prod(s_expansion[0], A_expansion[0]))
        else:
            final_expansion.append(high_order_expansion[n])
            n += 1
    return final_expansion

def inverse_transformation(forward_expansion, order, reverse_expansion=None, allow_pseudoinverse=False):
    if reverse_expansion is None:
        if allow_pseudoinverse and forward_expansion[0].shape[0] != forward_expansion[0].shape[1]:
            B = np.linalg.pinv(forward_expansion[0])
        else:
            B = np.linalg.inv(forward_expansion[0])
        reverse_expansion = [B]
    else:
        reverse_expansion = list(reverse_expansion)
        B = reverse_expansion[0]

    if not is_numeric(order): order = np.max(order)
    order = list(range(2, order+1))
    shared = B.ndim - 2

    for o in order:
        new_B = -tensor_reexpand(reverse_expansion, forward_expansion, [o])[-1]
        if isinstance(new_B, np.ndarray):
            # need to multiply in the inverse now, too
            new_B = vec_ops.vec_tensordot(new_B, B, axes=[-1, shared], shared=shared)
            # for _ in range(new_B.ndim - 1):
            #     new_B = np.tensordot(B, new_B, axes=[1, -2])
        reverse_expansion = reverse_expansion + [new_B]
    return reverse_expansion

def kron_prod_derivs(A_expansion, B_expansion, order):
    # s = A_expansion[0].ndim - 2
    C_expansion = tensorprod_deriv(A_expansion, B_expansion, order,
                                   axes=[[-1, -2], [-1, -2]]
                                   )
    # now we need to reshape based on the fact that we end up with nnxmm
    return [
        c.reshape(c.shape[:-4] + (c.shape[-4]*c.shape[-3], c.shape[-2]*c.shape[-1]))
        for c in C_expansion
    ]

def kron_sum_derivs(A_expansion, B_expansion, order):
    base_shape = B_expansion[0].shape[:-2]
    ns = len(base_shape)
    m = B_expansion[0].shape[-1]
    if ns > 0:
        I_expansion = np.expand_dims(np.eye(m), list(range(ns)))
        I_expansion = np.broadcast_to(I_expansion, base_shape + (m, m))
    else:
        I_expansion = np.eye(m)
    I_expansion = [I_expansion]
    left_expansion = kron_prod_derivs(A_expansion, I_expansion, order)

    n = A_expansion[0].shape[-1]
    if ns > 0:
        I_expansion = np.expand_dims(np.eye(n), list(range(ns)))
        I_expansion = np.broadcast_to(I_expansion, base_shape + (n, n))
    else:
        I_expansion = np.eye(n)
    I_expansion = [I_expansion]
    right_expansion = kron_prod_derivs(I_expansion, B_expansion, order)

    return [l+r for l,r in zip(left_expansion, right_expansion)]

def sylv_derivs(A_expansion, B_expansion, C_expansion, order, shared=None, ks_expansion=None, inv_expansion=None):
    n = A_expansion[0].shape[-1]
    m = B_expansion[0].shape[-1]
    B_expansion = [np.moveaxis(b, -1, -2) for b in B_expansion]
    if ks_expansion is not None:
        ks_expansion = list(ks_expansion) + kron_sum_derivs(A_expansion, B_expansion, order)
    else:
        ks_expansion = kron_sum_derivs(A_expansion, B_expansion, order)
    if isinstance(order, list):
        inv_expansion = matinv_deriv(ks_expansion, order[0], base_expansion=inv_expansion)
    else:
        inv_expansion = matinv_deriv(ks_expansion, order, base_expansion=inv_expansion)
    C_expansion = [np.reshape(c, c.shape[:-2] + (c.shape[-1] * c.shape[-2],)) for c in C_expansion]
    X_expansion = tensordot_deriv(C_expansion, inv_expansion, order,
                                  axes=[-1, -1],
                                  shared=shared
                                  )

    return [np.reshape(x, x.shape[:-1] + (n,m)) for x in X_expansion], (ks_expansion, inv_expansion)

def matsqrt_deriv(A_expansion, order):
    shared = A_expansion[0].ndim - 2
    if shared == 0: shared = None
    B_expansion = [
        vec_ops.frac_powh(A_expansion[0], 1/2, pow=lambda evals,_:np.sqrt(evals))
        # sqrt maybe faster than pow(..., 1/2)
    ]
    inv_expansion = None
    ks_expansion = None
    for o in range(1, order+1):
        new_expansion, (ks_expansion, inv_expansion) = sylv_derivs(B_expansion, B_expansion, A_expansion[1:],
                                                                   ks_expansion=ks_expansion,
                                                                   inv_expansion=inv_expansion,
                                                                   shared=shared,
                                                                   order=[o - 1])
        B_expansion = B_expansion + new_expansion

    return B_expansion

def matinv_deriv(A_expansion, order, base_expansion=None):
    shared = s = A_expansion[0].ndim - 2
    if shared == 0: shared = None
    if base_expansion is None:
        B = np.linalg.inv(A_expansion[0])
        base_expansion = [B]
    inverse_expansion = list(base_expansion)

    for o in range(len(inverse_expansion), order+1):
        expansion = tensorops_deriv(
            A_expansion[1:],
                [s, s + 1],
            inverse_expansion,
                [s, s + 0],
            inverse_expansion,
            order=[o - 1],
            shared=shared
        )
        inverse_expansion = inverse_expansion + [-expansion[-1]]
    return inverse_expansion

def matdet_deriv(forward_expansion, order):
    shared = forward_expansion[0].ndim - 2
    reverse_expansion = matinv_deriv(forward_expansion, order)
    tr_inner_exp = tensordot_deriv(forward_expansion[1:], reverse_expansion,
                                   order - 1,
                                   axes=[1+shared, 1+shared],
                                   shared=shared
                                   )
    tr_exp = [np.trace(a, axis1=-2, axis2=-1) for a in tr_inner_exp]
    det_exp = [np.linalg.det(forward_expansion[0])]
    for o in range(0, order):
        det_exp = det_exp + [scalarprod_deriv(det_exp, tr_exp, [o])[-1]]
    return det_exp

def _broadcast_mul(scalar, other):
    if isinstance(scalar, np.ndarray) and scalar.ndim < other.ndim:
        return np.expand_dims(scalar, list(range(scalar.ndim, other.ndim))) * other
    else:
        return scalar * other
def _scalarinv_deriv(scalar_expansion, o):
    from ..Combinatorics import IntegerPartitioner

    shared = scalar_expansion[0].ndim
    term = 0
    for parts in IntegerPartitioner.partitions(o):
        l = len(parts[0])
        scaling = ((-1) ** l) * math.factorial(l) / (scalar_expansion[0] ** (l + 1))
        term += sum(
            _broadcast_mul(scaling, nca_partition_prod(p, scalar_expansion[1:], shared=shared))
            for p in parts
        )
    return term

def scalarinv_deriv(scalar_expansion, order):
    return _deriv_construct(
        lambda : 1/scalar_expansion[0],
        lambda ords: [_scalarinv_deriv(scalar_expansion, o) for o in ords],
        order
    )

def _scalarpow_deriv(scalar_expansion, exp, o):
    from ..Combinatorics import IntegerPartitioner
    shared = scalar_expansion[0].ndim

    term = 0
    for parts in IntegerPartitioner.partitions(o):
        l = len(parts[0])
        factorial_terms = np.prod(exp - np.arange(l))
        if factorial_terms == 0:
            continue
        scaling = factorial_terms * (scalar_expansion[0] ** (exp - l))
        term += sum(
            _broadcast_mul(scaling, nca_partition_prod(p, scalar_expansion[1:], shared=shared))
            for p in parts
        )
    return term

def scalarpow_deriv(scalar_expansion, exp, order):
    return _deriv_construct(
        lambda : np.power(scalar_expansion[0], exp),
        lambda ords: [_scalarpow_deriv(scalar_expansion, exp, o) for o in ords],
        order
    )

def vec_norm_unit_deriv(vec_expansion, order, base_expansion=None):
    if base_expansion is None:
        a = vec_expansion[0]
        r = np.linalg.norm(a, axis=-1)
        if a.ndim > 1:
            u = a / r[..., np.newaxis]
        else:
            u = a / r
        base_expansion = [r, u]
    base_expansion = list(base_expansion)
    norm_inv_expansion = scalarinv_deriv(base_expansion, order=len(base_expansion)-2)

    a_expansion = [vec_expansion[0], vec_ops.identity_tensors(base_expansion[1].shape[:-1], 3)]

    if not is_numeric(order):
        order = max(order)

    for o in range(len(base_expansion), order+2):
        # if len(base_expansion) <= o:
        norm_inv_expansion = norm_inv_expansion + scalarinv_deriv(base_expansion, order=[o-1])
        base_expansion = base_expansion + scalarprod_deriv(norm_inv_expansion, a_expansion, order=[o-1])
        #     norm_expansion = norm_expansion + tensordot_deriv(vec_expansion, unit_expansion, axes=[-1, -1], order=[o],
        #                                                       shared=vec_expansion[0].ndim-1)
        #     norm_inv_expansion = norm_inv_expansion + scalarinv_deriv(norm_expansion, order=[o])
        # if len(unit_expansion) <= o:
        #     deriv_term = scalarprod_deriv(norm_inv_expansion, vec_expansion, order=[o])
        #     unit_expansion = unit_expansion + deriv_term

    # reexpand in terms of original vectors
    # raise Exception(vec_expansion)
    norm_expansion = [base_expansion[0]] + tensor_reexpand(
        vec_expansion[1:],
        base_expansion[1:],
        order + 1
        # shared=vec_expansion[0].ndim - 1
    )

    unit_expansion = [base_expansion[1]] + tensor_reexpand(
        vec_expansion[1:],
        base_expansion[2:],
        order
        # shared=vec_expansion[0].ndim - 1
    )

    return norm_expansion, unit_expansion

def vec_anglecos_deriv(A_expansion, B_expansion, order, unitized=False):
    if not unitized:
        A_expansion = vec_norm_unit_deriv(A_expansion, order)[1]
        B_expansion = vec_norm_unit_deriv(B_expansion, order)[1]
    shared = A_expansion[0].ndim - 1
    return tensordot_deriv(A_expansion, B_expansion, order=order, axes=[-1, -1], shared=shared)

def vec_cross_deriv(A_expansion, B_expansion, order):
    shared = A_expansion[0].ndim - 1
    base_shape = A_expansion[0].shape[:-1]
    e3 = np.broadcast_to(
        np.expand_dims(levi_cevita3, list(range(shared))),
        base_shape + (3, 3, 3)
    )
    return tensorops_deriv(
        B_expansion,
            [-1, shared],
        [e3],
            [-1, -1],
        A_expansion,
        order=order,
        shared=shared
    )

def vec_parallel_cross_norm_deriv(A_expansion, B_expansion, order):
    i3 = np.broadcast_to(np.eye(3)[np.newaxis], (3, 3, 3)).copy()
    i3[0, 0, 0] = 0
    i3[1, 1, 1] = 0
    i3[2, 2, 2] = 0

    base_shape = A_expansion[0].shape[:-1]
    shared = len(base_shape)
    i3 = np.broadcast_to(np.expand_dims(i3, list(range(shared))), base_shape + (3, 3, 3))

    pseudo_norm_2 = np.sqrt(tensorops_deriv(
        B_expansion,
            [-1, -1],
        [i3],
            [-1, -1],
        A_expansion,
        order=0,
        shared=shared
    )[0])

    if is_numeric(order):
        order = np.arange(order+1)

    A_exp = [
        vec_ops.vec_tensordot(A_expansion[o], pseudo_norm_2, axes=[-1, -1], shared=shared)
            if len(A_expansion) > o and not is_zero(A_expansion[o])  else
        0
        for o in order
    ]
    B_exp = [
        vec_ops.vec_tensordot(B_expansion[o], pseudo_norm_2, axes=[-1, -1], shared=shared)
            if len(B_expansion) > o and not is_zero(B_expansion[o]) else
        0
        for o in order
    ]
    pseudonorm_expansion = [a-b for a,b in zip(A_exp, B_exp)]
    return pseudonorm_expansion

def vec_anglesin_deriv(A_expansion, B_expansion, order, unitized=False, return_unit_vectors=True, planar=None, planar_threshold=1e-14):
    if is_numeric(order):
        max_order = order
    else:
        max_order = max(order)
    if not unitized:
        A_expansion = vec_norm_unit_deriv(A_expansion, max_order)[1]
        B_expansion = vec_norm_unit_deriv(B_expansion, max_order)[1]

    if not planar:
        vec_cross = vec_cross_deriv(A_expansion, B_expansion, order=max_order)
        if planar is None:
            planar = np.linalg.norm(vec_cross[0], axis=-1) <= planar_threshold
    else:
        vec_cross = None

    if np.all(planar):
        expansion = vec_parallel_cross_norm_deriv(A_expansion, B_expansion, order)
        if return_unit_vectors:
            return expansion, A_expansion
        else:
            return expansion
    elif np.any(planar):
        # gotta compute planar and nonplanar separately then remerge the tensors

        base_shape = A_expansion[0].shape[:-1]
        planar_pos = np.where(planar)
        nonplanar_pos = np.where(np.logical_not(planar))
        planar_A = [a[planar_pos] for a in A_expansion]
        planar_B = [b[planar_pos] for b in B_expansion]
        expansion = vec_parallel_cross_norm_deriv(planar_A, planar_B, order)
        planar_units = planar_A
        nonplanar_geoms = [v[nonplanar_pos] for v in vec_cross]
        norms, units = vec_norm_unit_deriv(nonplanar_geoms, order)

        final_tensors = [
            np.zeros(base_shape + e.shape[1:])
            for e in expansion
        ]
        for e,n,f in zip(expansion, norms, final_tensors):
            f[planar_pos] = e[planar_pos]
            f[nonplanar_pos] = n[nonplanar_pos]
        if return_unit_vectors:
            final_units = [
                np.zeros(base_shape + u.shape[1:])
                for u in units
            ]
            for e, n, f in zip(planar_units, units, final_units):
                f[planar_pos] = e[planar_pos]
                f[nonplanar_pos] = n[nonplanar_pos]

            return final_tensors, final_units
        else:
            return final_tensors
    else:
        norms, units = vec_norm_unit_deriv(vec_cross, order)
        if return_unit_vectors:
            return norms, units
        else:
            return norms

def vec_angle_deriv(A_expansion, B_expansion, order, unitized=False):
    if not unitized:
        units_A, A_expansion = vec_norm_unit_deriv(A_expansion, order)
        units_B, B_expansion = vec_norm_unit_deriv(B_expansion, order)
    cos_expansion = vec_anglecos_deriv(A_expansion, B_expansion, order, unitized=True)
    sin_expansion = vec_anglesin_deriv(A_expansion, B_expansion, order, unitized=True, return_unit_vectors=False)
    cos_sin_expansion = [
        np.moveaxis(np.array([c, s]), 0, -1)
        for c, s in zip(cos_expansion, sin_expansion)
    ]
    arctan_expansion = [
        np.array([
            -sin_expansion[0],
            cos_expansion[0]
        ]),
        np.array([
            [np.zeros_like(sin_expansion[0]), -np.ones_like(sin_expansion[0])],
            [np.ones_like(sin_expansion[0]),   np.zeros_like(cos_expansion[0])]
        ])
    ]
    # raise Exception(arctan_expansion)
    new_expansions = []
    for i, e in enumerate(arctan_expansion):
        for _ in range(i + 1):
            e = np.moveaxis(e, 0, -1)
        new_expansions.append(e)
    arctan_expansion = new_expansions
    angle_derivs = [np.arctan2(sin_expansion[0], cos_expansion[0])] + tensor_reexpand(
        cos_sin_expansion[1:],
        arctan_expansion,
        # axes=[-1, -1],
        order=order,
        # shared=arctan_expansion[0].ndim - 1
    )
    return angle_derivs

def vec_dihed_deriv(A_expansion, B_expansion, C_expansion, order, planar=None, planar_threshold=1e-14):
    # quick check

    if is_numeric(order):
        max_order = order
    else:
        max_order = max(order)

    axb_expansion = vec_cross_deriv(A_expansion, B_expansion, max_order)
    bxc_expansion = vec_cross_deriv(B_expansion, C_expansion, max_order)

    return vec_angle_deriv(axb_expansion, bxc_expansion, order, unitized=False)

def vec_plane_angle_deriv(A_expansion, B_expansion, C_expansion, D_expansion, order, planar=None, planar_threshold=1e-14):
    # quick check

    if is_numeric(order):
        max_order = order
    else:
        max_order = max(order)

    axb_expansion = vec_cross_deriv(A_expansion, B_expansion, max_order)
    cxd_expansion = vec_cross_deriv(C_expansion, D_expansion, max_order)

    return vec_angle_deriv(axb_expansion, cxd_expansion, order, unitized=False)


def nca_partition_terms(partition):
    """
    Computes the number of permutations for the non-commutative operation
    corresponding to the given partition

    :param cls:
    :param partition:
    :return:
    """
    _, counts = np.unique(partition, return_counts=True)

    # compute the reduced multinomial coefficient in stable (enough) form
    multinomial_num = np.flip(np.arange(1, 1 + np.sum(partition)))
    multinomial_denum = np.concatenate([
        np.arange(1, 1 + j)
        for j in partition
    ])
    multi_redux_terms = np.concatenate([
        np.arange(1, 1 + j)
        for j in counts
    ])
    full_denom = np.flip(np.sort(np.concatenate([multinomial_denum, multi_redux_terms])))
    numl = len(multinomial_num)
    denl = len(full_denom)
    if numl > denl:
        full_denom = np.concatenate([full_denom, np.ones(numl - denl, dtype=full_denom.dtype)])
    elif denl > numl:
        multinomial_num = np.concatenate([multinomial_num, np.ones(denl - numl, dtype=multinomial_num.dtype)])

    return np.prod(multinomial_num / full_denom)

def nca_symmetrize(tensor, partition,
                   shared=None,
                   identical=True,
                   reweight=None):
    from ..Combinatorics import UniquePermutations
    perm_counter = len(partition)
    perm_idx = []  # to establish the set of necessary permutations to make things symmetric
    for i in partition:
        if (not identical) or i > 1:
            perm_idx.extend([perm_counter] * i)
            perm_counter -= 1
        else:
            perm_idx.append(perm_counter)

    # sometimes we overcount, so we factor that out here
    perm_inds, _ = UniquePermutations(perm_idx).permutations(return_indices=True)
    if reweight or (reweight is None and identical):
        nterms = nca_partition_terms(partition)
        overcount = len(perm_inds) / nterms
        tensor = tensor / overcount

    if shared is None:
        perm_inds = [
            list(p) + list(range(len(p), tensor.ndim))
            for p in perm_inds
        ]
    else:
        l = list(range(shared))
        perm_inds = [
            l + [shared + pp for pp in p] + list(range(shared+len(p), tensor.ndim))
            for p in perm_inds
        ]

    return sum(
        tensor.transpose(p)
        for p in perm_inds
    )

def nca_partition_dot(partition, A_expansion, B_expansion, axes=None, shared=None, symmetrize=True):
    if axes is None:
        axes = [-1, (0 if shared is None else shared)]
    if len(B_expansion) <= len(partition) - 1:
        return 0
    B = B_expansion[len(partition) - 1]
    if is_numeric(B) and B == 0:
        return 0
    a_ax, b_ax = [[x] if is_numeric(x) else x for x in axes]
    b_ax = [B.ndim + b if b < 0 else b for b in b_ax]
    for i in reversed(partition):
        if len(A_expansion) <= i - 1:
            return 0
        A = A_expansion[i - 1]
        if is_numeric(A) and A == 0:
            return 0
        if shared is None:
            B = np.tensordot(A, B, axes=[a_ax, b_ax])
        else:
            B = vec_ops.vec_tensordot(A, B, axes=[a_ax, b_ax], shared=shared)
        b_ax = [min(B.ndim - 1, b + A.ndim - 1 - shared) for b in b_ax]

    if symmetrize:
        B = nca_symmetrize(B, partition, shared=shared)
    return B

def nca_partition_prod(partition, A_expansion, shared=None, symmetrize=True):
    i = partition[0]
    if len(A_expansion) <= i - 1:
        return 0
    A = A_expansion[i - 1]
    if is_numeric(A) and A == 0:
        return 0

    B = A
    for i in partition[1:]:
        if len(A_expansion) <= i - 1:
            return 0
        A = A_expansion[i - 1]
        if is_numeric(A) and A == 0:
            return 0
        if shared is not None:
            axes = [list(range(shared, A.ndim)), list(range(shared, B.ndim))]
        else:
            axes = [list(range(A.ndim)), list(range(B.ndim))]
        B = vec_ops.vec_outer(A, B, axes=axes)

    if symmetrize:
        B = nca_symmetrize(B, partition, shared=shared)
    return B

def tensor_reexpand(derivs, vals, order=None, axes=None):
    from ..Combinatorics import IntegerPartitioner
    terms = []
    if order is None:
        order = len(vals)

    shared = derivs[0].ndim - 2
    if is_numeric(order): order = list(range(1, order+1))

    for o in order:
        term = sum(
            nca_partition_dot(p, derivs, vals, axes=axes, shared=shared)
            for parts in IntegerPartitioner.partitions(o)
            for p in parts
        )
        terms.append(term)

    return terms

def optimizing_transformation(expansion, order):
    V = expansion

    zero_order = 0
    for v in V:
        if is_zero(v):
            zero_order += 1
        else:
            break

    if zero_order > 2:
        raise ValueError("can't optimize without gradient or Hessian (tensor inverses not implemented)")

    W = V[zero_order]
    if zero_order == 0:
        w = W
        Q = [np.eye(V[1].shape[0])]
    elif zero_order == 1 and np.allclose(np.diag(np.diag(W)), W):
        Q = [np.eye(V[1].shape[0])]
        w = np.diag(W)
    else:
        Q = [np.linalg.inv(W)]
        w = np.ones(len(W))

    for o in range(1, order + 1):
        V_rem = tensor_reexpand(Q, V, [o], axes=[-1, 0])[-1]
        w = w[np.newaxis]
        new_Q = -V_rem / ((o + 2) * w)
        Q = Q + [new_Q]

    return Q

def apply_nca_multi_ops(partition, expansions, ops, shared):
    terms = [
        e[p] if len(e) > p else 0
        for e,p in zip(expansions, partition)
    ]
    if any(is_numeric(t) and t == 0 for t in terms): return 0

    if shared is None: shared = 0

    A = terms[0]
    # d = partition[0]
    d = A.ndim - 2 - shared
    scaling = 2 if len(partition) == 3 and tuple(partition) == (1, 0, 1) else 1
    for B,p,(op, axes, contract) in zip(terms[1:], partition[1:], ops):
        a_ax, b_ax = [[x] if is_numeric(x) else x for x in axes]
        a_ax = [a if a >= 0 else (A.ndim + a - d) for a in a_ax]
        b_ax = [b if b >= 0 else (B.ndim + b - p) for b in b_ax]
        if contract: # axes disappear, so we just account for the shifts
            axes = [[x+d for x in a_ax], [x+p for x in b_ax]]
            deriv_axis = A.ndim - len(a_ax)
        else: # axes appeared, so we need to include those in the product
              # we _forced_ axes to be at the end of the arrays since it was too hard
              # to keep track of them otherwise...
              # actually I guess I could have put the derivative axes at the end...
              # and then the axes would never change...but that has other complications
            axes = [
                [shared + i for i in range(d)] + [x+d for x in a_ax],
                [shared + i for i in range(p)] + [x+p for x in b_ax]
            ]
            deriv_axis = A.ndim
        if shared == 0:
            A = op(A, B, axes=axes)
        else:
            A = op(A, B, axes=axes, shared=shared)
        # next we need to move all of the derivative axes in the second tensor to the beginning
        # so we can symmetrize

        for i in range(p):
            A = np.moveaxis(A, deriv_axis+i, shared)
        d += B.ndim - 2 - shared
    partition = [p for p in partition if p > 0]
    if len(partition) > 1:
        A = nca_symmetrize(A, partition, shared=shared, identical=False)
    return A
def nca_multi_op_order_deriv(partition_generator, order, expansions, ops, shared):
    full = None
    for part in partition_generator.get_terms(order):
        term = apply_nca_multi_ops(part, expansions, ops, shared)
        if full is None:
            full = term
        else:
            full = full + term
    return full

def nca_canonicalize_multiops(expansion_op_pairs):
    if len(expansion_op_pairs) % 2 == 0: raise ValueError("invalid number of operations/expansions")

    expansions = []
    ops = []
    for i, t in enumerate(expansion_op_pairs):
        if i % 2 == 0:
            expansions.append([np.asanyarray(A) for A in t])
        else:
            op = None
            axes = None
            contract = None
            if isinstance(t, dict):
                op = t.get('op', None)
                axes = t.get('axes', None)
                contract = t.get('contract', None)
            elif isinstance(t, str):
                op = t
            elif is_numeric(t[0]) or is_numeric(t[0][0]):
                axes = t
            else:
                op = t[0]
                if len(t) > 1: axes = t[1]
                if len(t) > 2: contract = t[2]

            if op is None and axes is not None:
                op = '.'

            if isinstance(op, str):
                if op in {'.', 'dot', 'contract', 'inner'}:
                    op = _contract
                    if axes is None: axes = [-1, 0]
                    if contract is None: contract = True
                elif op in {'x', 'prod', 'product', 'outer'}:
                    op = _product
                    # if axes is None: axes = None
                    if contract is None: contract = False
                else:
                    raise ValueError(f"can't canonicalize operation {op}")

            if contract is None:
                contract = op is _contract

            ops.append([op, axes, contract])

    return expansions, ops

def _contract(a, b, axes=None, shared=None):
    if shared is not None:
        res = vec_ops.vec_tensordot(a, b, axes=axes, shared=shared)
    else:
        res = np.tensordot(a, b, axes=axes)
    return res
def _product(a, b, axes=None, shared=None):
    return vec_ops.vec_outer(a, b, axes=axes)

def tensorops_deriv(
        *expansion_op_pairs,
        order,
        shared=None
):
    from ..Combinatorics import SymmetricGroupGenerator
    expansions, ops = nca_canonicalize_multiops(expansion_op_pairs)

    if isinstance(order, int):
        order = list(range(order+1))

    partition_generator = SymmetricGroupGenerator(len(expansions))
    derivs = [
        nca_multi_op_order_deriv(partition_generator, o, expansions, ops, shared)
        for o in order
    ]

    return derivs


from distutils.core import setup, Extension
import shutil, os, sys

# from ..config import CPotentialCompilerConfig as cpcc

lib_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
src_dir = os.path.join(lib_dir, "src")
libname = "ZachLib"


def get_macros():
    macs = []
    # if cpcc["UseMPI"]:
    #     macs.append(("USE_MPI", None))
    return macs


def get_extension():
    return Extension(
        libname,
        sources=[libname + '.cpp', "PyExtLib.cpp"],
        define_macros=get_macros()
    )


def setup_compile():
    return setup(
        name=libname,
        version='1.0',
        description='Performance critical parts of Zachary',
        ext_modules=[get_extension()],
        language="c++"
    )


def compile():
    curdir = os.getcwd()
    try:
        os.chdir(src_dir)

        sysargv1 = sys.argv
        sys.argv = ['build', 'build_ext', '--inplace']
        try:
            setup_compile()
        finally:
            sys.argv = sysargv1
    finally:
        os.chdir(curdir)


def find_source():
    ext = ""
    target = os.path.join(lib_dir, libname)
    src = None
    for f in os.listdir(os.path.join(lib_dir, "src")):
        if f.startswith(libname) and f.endswith(".so"):
            ext = ".so"
            src = os.path.join(lib_dir, "src", f)
            target += ext
            break
        elif f.startswith(libname) and f.endswith(".pyd"):
            ext = ".pyd"
            src = os.path.join(lib_dir, "src", f)
            target += ext
            break

    return src, target


def load():
    compile()
    src, target = find_source()
    if src is not None:
        try:
            os.remove(target)
        except:
            pass
        os.rename(src, target)
        shutil.rmtree(os.path.join(src_dir, "build"))
        return src
    else:
        return None


if __name__ == "__main__":
    loaded = load()
    if loaded is None:
        raise ImportError("{}: library didn't compile".format(libname))



"""
Provides some support for working with the python bindings for external programs, like OpenBabel
Mostly relevant for doing format conversions/parsing, but other utilities do exist.
"""

__all__ = []
from .ChemToolkits import *; from .ChemToolkits import __all__ as exposed
__all__ += exposed
from .Visualizers import *; from .Visualizers import __all__ as exposed
__all__ += exposed
from .RDKit import *; from .RDKit import __all__ as exposed
__all__ += exposed


"""
Implements all of the Taylor-series derived functionality in Zachary (i.e. all the finite-difference stuff and the function expansions)
"""

from .Derivatives import *
from .FunctionExpansions import *
from .FiniteDifferenceFunction import __all__ as FiniteDifferenceFunction__all__
from .FiniteDifferenceFunction import *

__all__ = (
    FiniteDifferenceFunction__all__ +
    FunctionExpansions.__all__ +
    Derivatives.__all__
)


"""
Provides a Loader object to load a potential from a C++ extension
"""

import os, numpy as np
import platform

from .. import CLoader, ModuleLoader
from .Module import FFIModule

__all__ = [
    "FFILoader"
]

class FFILoader:
    """
    Provides a standardized way to load and compile a potential using a potential template
    """
    __props__ = [
        "src_ext",
        "description",
        "version",
        "include_dirs",
        "linked_libs",
        "macros",
        "source_files",
        "build_script",
        "requires_make",
        "out_dir",
        "cleanup_build",
        'build_kwargs'
        'nodebug',
        'threaded',
        'extra_compile_args',
        'extra_link_args',
        'recompile'
    ]

    # src_folder = os.path.join(os.path.dirname(__file__), "src")
    libs_folder = os.path.join(os.path.dirname(__file__), "libs")
    cpp_std = '-std=c++17'
    def __init__(self,
                 name,
                 src=None,
                 src_ext='src',
                 load_path=None,
                 description="A compiled potential",
                 version="1.0.0",
                 include_dirs=None,
                 linked_libs=None,
                 runtime_dirs=None,
                 macros=None,
                 source_files=None,
                 build_script=None,
                 requires_make=True,
                 out_dir=None,
                 cleanup_build=True,
                 pointer_name=None,
                 build_kwargs=None,
                 nodebug=False,
                 threaded=False,
                 extra_compile_args=None,
                 extra_link_args=None,
                 recompile=False
                 ):
        # if python_potential is False:

        if include_dirs is None:
            include_dirs = []
        include_dirs = (
                               tuple(include_dirs)
                               + (self.libs_folder, np.get_include())
                               # + (PotentialCaller.TBB_CentOS, PotentialCaller.TBB_Ubutu)
        )
        if linked_libs is None:
            linked_libs = []
        linked_libs = tuple(linked_libs) #+ ("plzffi", )

        self.threaded = threaded
        if macros is None:
            macros = []
        if nodebug:
            macros = list(macros) + [('_NODEBUG', True)]
        threading_flags = (["-fopenmp"] if self.threaded else [])
        # if platform.system() == "Darwin":
        #     threading_flags = ['-Xprepocessor'] + threading_flags
        self.c_loader = CLoader(
            name,
            src,
            load_path=load_path,
            src_ext=src_ext,
            description=description,
            version=version,
            include_dirs=include_dirs,
            runtime_dirs=runtime_dirs,
            linked_libs=(("omp",) if self.threaded else ()) + linked_libs,
            macros=macros,
            source_files=source_files,
            build_script=build_script,
            requires_make=requires_make,
            out_dir=out_dir,
            cleanup_build=cleanup_build,
            extra_compile_args=threading_flags + [self.cpp_std] + ([] if extra_compile_args is None else list(extra_compile_args)),
            extra_link_args=[] if extra_link_args is None else list(extra_link_args),
            recompile=recompile,
            **({} if build_kwargs is None else build_kwargs)
        )
        # else:
        #     self.c_loader = None

        # Need to insert code here to allow for new caller API to work
        self._lib = None

        self._attr = pointer_name
        # self.function_name = pointer_name

    @property
    def lib(self):
        if self._lib is None:
            self._lib = self.c_loader.load()
        return self._lib
    @property
    def caller_api_version(self):
        if hasattr(self.lib, "_FFIModule"): # currently how we're dispatching
            return 2
        else:
            return 1
    @property
    def call_obj(self):
        """
        The object that defines how to call the potential.
        Can either be a pure python function, an FFIModule, or a PyCapsule

        :return:
        :rtype:
        """
        return FFIModule.from_module(self.lib)


from .CoordinateSystemConverter import CoordinateSystemConverter
from .CommonCoordinateSystems import CartesianCoordinates3D, SphericalCoordinates
from ...Numputils import *
import numpy as np

class SphericalToCartesianConverter(CoordinateSystemConverter):
    """
    A converter class for going from ZMatrix coordinates to Cartesian coordinates
    """

    @property
    def types(self):
        return (SphericalCoordinates, CartesianCoordinates3D)

    def convert_many(self,
                     coords,
                     origin=None,
                     axes=None,
                     use_rad=True,
                     **kw
                     ):
        """
        Expects to get a list of configurations
        These will look like:
            [
                [dist, angle, dihedral]
                ...
            ]

        **For efficiency it is assumed that all configurations have the same length**

        :param coords:
        :type coords:
        :param origins:
        :type origins:
        :param axes:
        :type axes:
        :param use_rad:
        :type use_rad:
        :param kw:
        :type kw:
        :param ordering:
        :type ordering:
        :param return_derivs:
        :type return_derivs:
        :return:
        :rtype:
        """

        coords = np.asanyarray(coords)
        if origin is None:
            origin = [0., 0., 0.]
        origin = np.asanyarray(origin)
        for d in range(coords.ndim - origin.ndim):
            origin = np.expand_dims(origin, 0)
        if axes is None:
            axes = [[1., 0., 0.], [0., 0., 1.]]
        axes = np.asanyarray(axes)
        for d in range(coords.ndim - axes[0].ndim):
            axes = [np.expand_dims(a, 0) for a in axes]

        r = coords[..., 0]
        a = coords[..., 1]
        if not use_rad:
            a = np.deg2rad(a)
        d = coords[..., 2]
        if not use_rad:
            d = np.deg2rad(d)
        pts = polar_to_cartesian(origin, axes[0], axes[1], r, a, d)

        return pts, dict(**kw, origin=origin, axes=axes)

    def convert(self, coords, **kw):
        """dipatches to convert_many but only pulls the first"""
        total_points, opts = self.convert_many(coords[np.newaxis], **kw)
        return total_points[0], opts

__converters__ = [ SphericalToCartesianConverter() ]


import numpy as np

from ..JHTML import JHTML, HTML, HTMLWidgets
# from ..Apps import WrapperComponent
# import uuid

__all__ = [
    # "D3API",
    "NGLAPI"
]
__reload_hook__ = ["..JHTML", "..Apps"]

class NGLAPI:
    _api_versions = {}
    @classmethod
    def load(cls, version='v5'):
        if version not in cls._api_versions:
            cls._api_versions[version] = JHTML.JavascriptAPI(
    ngl_init="""
let model = this.model;
if (typeof(model) === 'undefined' || model === null){
    model = this;
}

if (typeof(model.ngl) === 'undefined' || model.ngl === null) {
    return import("https://unpkg.com/ngl")
      .then(() => (model.ngl = ngl));
} else {
    return Promise.resolve();
}""",

    ngl_stage="""
let model = this.model;
if (typeof(model) === 'undefined' || model === null){
    model = this;
}

model.callHandler("ngl_init").then(()=>{
    let ngl = model.ngl;
    let stage = model.stage;
    if (typeof(stage) === 'undefined' || stage === null) {
        if (event.content.hasOwnProperty('id')) {
            stage = new ngl.Stage(event.content['id']);
            model.stage = stage;
        } else {
            alert("no id for stage")
        }
    }
    return model.stage
})
    """,

    ngl_load="""

let model = this.model;
if (typeof(model) === 'undefined' || model === null){
    model = this;
}

model.callHandler("ngl_stage").then((stage)=>{
    return stage.loadFile(type=)
})
    """,

    ngl_call="""
let model = this.model;
if (typeof(model) === 'undefined' || model === null){
    model = this;
}

model.callHandler("ngl_stage").then((stage)=>{
    let methods = event.content['methods'];
    let args = event.content['args'];
    let debug = event.content['debug'];
    
    if (debug) { console.log(stage); }
    return methods.reduce(
        (acc, method, index) => {
            if (method in stage) {
                let fn = stage[method];
                if (debug) { console.log(method, args[index], fn); }
                let argl = [svg, ...args[index]];
                fn.call(...argl);
            } else {
                if (debug) { console.log("no attr", method); }
            }
        }, 
        []
    )
})
""",

    ngl_props="""
let model = this.model;
if (typeof(model) === 'undefined' || model === null){
    model = this;
}

model.callHandler("d3_init").then(()=>{
    let d3 = model.d3;
    let props = event.content['properties'];
    let debug = event.content['debug'];
    let el = this.el;
    if (typeof(el) === 'undefined' || el === null) {
        if (event.content.hasOwnProperty('id')) {
            el = "#" + event.content['id']
        }
    }

    if (typeof(el) === 'undefined' || el === null) {
        alert("no d3 object provided");
    } else {
        let svg = d3.select(el);
        if (debug) { console.log(el, svg); }
        let attrs = props.map(
            (prop) => svg.attr(prop)
        );
        this.send({type:"content", content:attrs});
    }
})
"""
)
        return cls._api_versions[version]


"""
coeffuncs defines the functions we use when computing coefficients for FD
"""
import numpy as np, math

def StirlingS1(n):
    """Computes the Stirling numbers

    :param n:
    :type n:
    :return:
    :rtype:
    """
    stirlings = np.eye(n)
    for i in range(n):
        for j in range(i+1):
            stirlings[i, j] = (-1)**(i-j) *( (i-1)*abs(stirlings[i-1, j]) + abs(stirlings[i-1, j-1]))
    return stirlings

def Binomial(n):
    """

    :param n:
    :type n:
    :return:
    :rtype:
    """
    binomials = np.eye(n)
    binomials[:, 0] = 1
    for i in range(2, n):
        if i%2 == 0:
            k = i/2 + 1
        else:
            k = (i+1)/2
        for j in range(int(k)):
            binomials[i, j] = binomials[i, i-j] = binomials[i-1, j-1] + binomials[i-1, j]
    return binomials

# def _StirlingS1(n):
#     """simple recursive definition of the StirlingS1 function in Mathematica
#     implemented at the C level mostly just for fun
#
#     :param n:
#     :type n:
#     :return:
#     :rtype:
#     """
#
#     return _StirlingS1(n)
#
# def Binomial(n):
#     """simple recursive Binomial coefficients up to r, computed all at once to vectorize later ops
#     wastes space, justified by assuming a small-ish value for n
#
#     :param n:
#     :type n:
#     :return:
#     :rtype:
#     """
#     return _Binomial(n)

def GammaBinomial(s, n):
    """Generalized binomial gamma function

    :param s:
    :type s:
    :param n:
    :type n:
    :return:
    :rtype:
    """
    g = math.gamma
    g1 = g(s+1)
    g2 = np.array([g(m+1)*g(s-m+1) for m in range(n)])
    g3 = g1/g2
    return g3

def Factorial(n):
    """I was hoping to do this in some built in way with numpy...but I guess it's not possible?
    looks like by default things don't vectorize and just call math.factorial

    :param n:
    :type n:
    :return:
    :rtype:
    """

    base = np.arange(n, dtype=np.int64)
    base[0] = 1
    for i in range(1, n):
        base[i] = base[i]*base[i-1]
    return base

def EvenFiniteDifferenceWeights(m, s, n):
    """Finds the series coefficients for x^s*ln(x)^m centered at x=1. Uses the method:

             Table[
               Sum[
                ((-1)^(r - k))*Binomial[r, k]*
                    Binomial[s, r - j] StirlingS1[j, m] (m!/j!),
                {r, k, n},
                {j, 0, r}
                ],
               {k, 0, n}
               ]
             ]

        which is shown by J.M. here: https://chat.stackexchange.com/transcript/message/49528234#49528234

    :param m: the order of the derivative requested
    :type m:
    :param s: the offset of the point at which the derivative is requested from the left edge of the stencil
    :type s:
    :param n: the number of points used in the stencil
    :type n:
    :return:
    :rtype:
    """

    n = n+1 # in J.M.'s algorithm we go from 0 to n in Mathematica -- which means we have n+1 elements
    stirlings = StirlingS1(n)[:, m]
    bins = Binomial(n)
    sTest = s - int(s)
    if sTest == 0:
        bges = bins[ int(s) ]
    else:
        bges = GammaBinomial(s, n)
    bges = np.flip(bges)
    facs = Factorial(n)
    fcos = facs[m]/facs # factorial coefficient (m!/j!)

    coeffs = np.zeros(n)
    bs = bges
    ss = stirlings
    fs = fcos
    # print(n, m, s, fcos, bges)
    for k in range(n):
        # each of these bits here should go from
        # Binomial[s, r - j] * StirlingS1[j, m] *
        bits = np.zeros(n-k)
        for r in range(k+1, n+1):
            bits[r-k-1] = np.dot(bs[-r:], ss[:r]*fs[:r])

        # (-1)^(r - k))*Binomial[r, k]
        cs = (-1)**(np.arange(n-k)) * bins[k:n, k]
        # print(bits, file=sys.stderr)
        coeffs[k] = np.dot(cs, bits)

    # print(coeffs)

    return coeffs

# _UnevenFiniteDifferenceWeights = None
def UnevenFiniteDifferenceWeights(m, z, x):
    # global _UnevenFiniteDifferenceWeights
    # if _UnevenFiniteDifferenceWeights is None:
    #     try:
    #         from .lib import UnevenFiniteDifferenceWeights as _UnevenFiniteDifferenceWeights
    #     except ImportError:
    #         def _UnevenFiniteDifferenceWeights(m, s, n):
    #
    #             raise NotImplementedError("Need to translate the C implementation to pure python; don't want to use an extension for something so minimal")
    # return _UnevenFiniteDifferenceWeights(m, z, x)

    # We're gonna provide an inefficient imp, but this function is so cheap to start out that who cares


    # Py_ssize_t n = PyObject_Length(xArray) - 1;
    # int dims[2] = {n+1, m+1};
    # PyObject *cArray = _CreateArray(2, dims, "zeros");
    # double *c = _GetDataArray<double>(cArray);

    # this is basically just copied directly from the OG source
    n = len(x) - 1
    dims = [n+1, m+1]
    c = np.zeros(dims)
    dxProdOld = 1.
    dz = x[0] - z
    c[0, 0] = 1.
    for i in range(1, n+1):
        mn = min(i, m)
        dxProd = 1.
        dzOld = dz
        dz = x[i] - z
        for j in range(0, i):
            dx = x[i] - x[j]
            dxProd *= dx
            if j == i-1:
                for k in range(mn, 0, -1):
                    c2 = c[i-1, k-1]
                    c3 = c[i-1, k]
                    weight = (k*c2 - dzOld * c3)*dxProdOld/dxProd
                    c[i, k] = weight
                c[i, 0] = -dzOld*c[i-1, 0]*dxProdOld/dxProd

            for k in range(mn, 0, -1):
                weight = (dz*c[j, k] - k*c[j, k-1])/dx
                c[j, k] = weight
            weight = dz*c[j, 0]/dx
            c[j, 0] = weight
        dxProdOld = dxProd

    return c



import numpy as np
from collections import OrderedDict

__all__ = [
    "StructuredType",
    "StructuredTypeArray",
    "DisappearingType"
]

class StructuredType:
    """
    Represents a structured type with a defined calculus to simplify the construction of combined types when writing
    parsers that take multi-typed data

    Supports a compound StructuredType where the types are keyed
    """
    def __init__(self,
                 base_type,
                 shape = None,
                 is_alternative = False,
                 is_optional = False,
                 default_value = None
                 ):

        if shape is None and isinstance(base_type, tuple) and len(base_type) == 2 and (
                isinstance(base_type[1], int) or isinstance(base_type[1], tuple)
        ): # to make it possible to initialize the type like (str, 3) or (int, (3,))
            base_type, shape = base_type

        self.dtype = base_type
        self.shape = shape

        # there's some question as to how we should declare alternative types?
        # currently we have a flag in the type but we could also make 3 subtypes of StructuredType:
        #   AlternativeTypes(...), PrimitiveType(...), CompoundType(...) ?

        self.is_alternative = is_alternative
        self.is_optional = is_optional


        self.default = default_value if default_value is not None else self._infer_default_value() # for optional types

    def _infer_default_value(self):
        missing = None
        if isinstance(self.dtype, type):
            if self.dtype is str:
                missing = "[NaN]"
            elif issubclass(self.dtype, (int, np.integer)):
                missing = int(-121e12) # -(Na e N)
            elif issubclass(self.dtype, (float, np.floating)):
                missing = np.nan
        # I guess we're gonna assume no other cases will realistically need padding/defaults?
        return missing


    @property
    def is_simple(self):
        return (
                self.dtype is None or (isinstance(self.dtype, type) and not self.is_optional and not self.is_alternative)
        )

    def add_types(self, other):
        """
        Constructs a new type by treating the two objects as siblings, that is if they can be merged due to type and
        shape similarity they will be, otherwise a non-nesting structure will be constructed from them

        We'll also want a nesting version of this I'm guessing, which probably we hook into __call__

        :param other:
        :type other:
        :return:
        :rtype:
        """

        if other is DisappearingType or isinstance(other, type(DisappearingType)):
            return self
        elif self is DisappearingType or isinstance(self, type(DisappearingType)):
            return other
        else:
            if not isinstance(other, StructuredType):
                other = StructuredType(other)

            # we define a very simple calculus on the structured types
            #   if both are the same we simply futz with the shapes if possible
            #   if they're different we construct a compound type from the two structured types
            # _unless_ we have an alternative type in which case...?

            if self.shape is None or other.shape is None:
                shape_mismatches = None
            else:
                shape_mismatches = np.array([(x is None) or (y is None) or int(x != y) for x,y in zip(self.shape, other.shape)])
            if (
                    self.is_simple and
                    self.dtype == other.dtype and (
                            shape_mismatches is None or (
                            len(self.shape) == len(other.shape) and np.sum(shape_mismatches) <= 1
                            )
                    )
                ):

                if self.shape is None and other.shape is None:
                    new_shape = (2,)
                elif self.shape is None:
                    new_shape = other.shape[:-1] + (other.shape[-1]+1, ) # this isn't quite right but will work for now
                elif other.shape is None:
                    new_shape = self.shape[:-1] + (self.shape[-1]+1, ) # this isn't quite right but will work for now
                else:
                    mismatch_pos = np.argwhere(shape_mismatches == 1)
                    if len(mismatch_pos) == 0:
                        mismatch_pos = -1
                    else:
                        mismatch_pos = mismatch_pos[-1]
                    new_shape = list(self.shape)
                    me = self.shape[mismatch_pos]
                    yu = other.shape[mismatch_pos]
                    if me is None or yu is None:
                        new_shape[mismatch_pos] = None
                    else:
                        new_shape[mismatch_pos] = me + yu

                import copy
                new = copy.copy(self)
                new.shape = tuple(new_shape)
                return new
            elif isinstance(self.dtype, type):
                # we do this because who knows what to do here otherwise...?
                return type(self)((self, other))
            else:
                import copy

                new = copy.copy(self)
                new.dtype += (other,)
                return new

    def __add__(self, other):
        return self.add_types(other)

    def compound_types(self, other):
        """Creates a structured type where rather than merging types they simply compound onto one another

        :param other:
        :type other:
        :return:
        :rtype:
        """
        raise NotImplemented
    def __call__(self, other):
        return self.compound_types(other)


    def repeat(self, n = None, m = None):
        """Returns a new version of the type, but with the appropriate shape for being repeated n-to-m times

        :param n:
        :type n:
        :param m:
        :type m:
        :return:
        :rtype:
        """

        import copy
        new = copy.copy(self)
        if new.shape is None:
            new.shape = (None,) if m is None else (m,)
        # elif n is not None and m is not None and n == m:
        #     new.shape = (n,) + new.shape
        elif m is not None: # if we have any kind of max number of elements, we know the size of the array we need to allocate
            new.shape = (m,) + new.shape
        else:
            new.shape = (None,) + new.shape # we won't handle things like between 10 and 20 elements for now
        return new

    def drop_axis(self, axis = 0):
        """Returns a new version of the type, but with the appropriate shape for dropping an axis

        :param axis:
        :type axis: int
        :return:
        :rtype:
        """

        import copy
        new = copy.copy(self)
        if new.shape is not None:
            new.shape = new.shape[:axis] + new.shape[axis+1:]
        elif not new.is_simple:
            # means we need to strip the axis from the subtypes
            dt = new.dtype
            if isinstance(dt, OrderedDict):
                new.dtype = OrderedDict( (k, d.drop_axis(axis=axis)) for k,d in dt.items())
            else:
                new.dtype = tuple(d.drop_axis(axis=axis) for d in dt)
        return new

    def extend_shape(self, base_shape):
        """Extends the shape of the type such that base_shape precedes the existing shape

        :param base_shape:
        :type base_shape:
        :return:
        :rtype:
        """
        shape = self.shape
        if self.shape is None:
            self.shape = base_shape
        else:
            self.shape = base_shape + shape

    def _condense_types(self, base_types, shape):
        # what do I do about this existing shape...?

        raw_types = [ b.dtype for b in base_types ]
        shapes = [ b.shape for b in base_types ]

        if shape is None and (
                all(r is raw_types[0] for r in raw_types) and
                (
                        all( s is None for s in shapes ) or
                        all( isinstance(s, tuple) and s == shapes[0] for s in shapes )
                )
        ):
            # means we can actually condense them but I might want to be a bit smarter with how I handle the tuple shape check...

            base_types = base_types[0] # no differences in type
            if shapes[0] is None:
                shape = (len(shapes),)
            else:
                shape += (len(shapes),)

        return base_types, shape


    def __repr__(self):
        return "{}({}, shape={})".format(
            type(self).__name__,
            self.dtype,
            self.shape
        )

class DisappearingTypeClass(StructuredType):
    """
    A special type that is entirely ignored in the structured type algebra
    """
    def __init__(self):
        self.is_disappearing = True
        super().__init__(None)
DisappearingType = DisappearingTypeClass() # redefinition but it should be a singleton anyway
DisappearingType.__name__ = 'DisappearingType'


"""
Something to think about w.r.t the correspondance between the stype shape and the stated shape of the 
array. 

There are two possible setups, really, in the first setup we have two cases:
    Simple stype => no shape for the stype
    Compound stype => shape of array is one _larger_ than shape of stype

In the second case:
    Simple dtype => no shape for the stype
    Compound dtype => shape of array is the shape of the stype

It's not clear that either is better than the other, but whichever one is chose, we need to be
consistent throughout the code in how we work with it.
"""

class StructuredTypeArray:
    """
    Represents an array of objects defined by the StructuredType spec provided
    mostly useful as it dispatches to NumPy where things are simple enough to do so

    It has a system to dispatch intelligently based on the type of array provided
    The kinds of structures supported are: OrderedDict, list, and np.ndarray

    A _simple_ StructuredTypeArray is one that can just be represented as a single np.ndarray
    A _compound_ StructuredTypeArray requires either a list or OrderedDict of StructuredTypeArray subarrays
    """

    # at some point this should make use of the more complex structured dtypes that NumPy provides...
    # for now we'll stick with this format, but using more NumPy will make stuff more efficient and easier to post-process
    def __init__(self, stype, num_elements = 50, padding_mode = 'fill', padding_value = None):
        """
        :param stype:
        :type stype: StructuredType
        :param num_elements: number of default elements in dynamically sized arrays
        :type num_elements: int
        """

        if not isinstance(stype, StructuredType):
            stype = StructuredType(stype)

        self._is_simple = stype.is_simple
        self._extend_along = None
        self._dtype = None
        self._stype = stype

        self._filled_to = None

        self._default_num_elements = num_elements
        self._array = None
        self._array = self.empty_array() # empty_array tries to use shape if possible
        self._append_depth = -1

        self.padding_mode = padding_mode
        self.padding_value = stype.default

    @property
    def is_simple(self):
        """Just returns wheter the core datatype is simple

        :return:
        :rtype:
        """
        return self._is_simple
    @property
    def dict_like(self):
        return isinstance(self._array, (dict, OrderedDict))
    @property
    def extension_axis(self):
        """Determines which axis to extend when adding more memory to the array
        :return:
        :rtype:
        """
        if self._extend_along is None:
            if self._stype.shape is None:
                self._extend_along = 0
            else:
                shape_nones = np.array([x is None for x in self._stype.shape], dtype=bool)
                if len(shape_nones) == 0:
                    self._extend_along = 0
                else:
                    self._extend_along = np.arange(len(shape_nones))
                    hmm = self._extend_along[shape_nones]
                    if len(hmm) == 0:
                        hmm = self._extend_along
                    self._extend_along = hmm[0]
        return self._extend_along
    @extension_axis.setter
    def extension_axis(self, ax):
        self._extend_along = ax
    @property
    def shape(self):
        if self._array is None:
            return None
        if self.is_simple:
            if isinstance(self._array, np.ndarray):
                # self._shape = list(self._array.shape)
                # self._shape[self.extension_axis] = self.filled_to[self.extension_axis] # this will mess up on things like (3,)...
                # self._shape = tuple(self._shape)
                self._shape = tuple(self.filled_to) if self.filled_to is not None else None
            else:
                self._shape = None
        else:
            if isinstance(self._array, OrderedDict):
                self._shape = [s.shape for s in self._array.values()]
            else:
                self._shape = [s.shape for s in self._array]
        return self._shape
    @shape.setter
    def shape(self, s):
        self._shape = s
    @property
    def block_size(self):
        if self.is_simple:
            s = list(self.shape)
            s[self.extension_axis] = 1 # this will mess up on things like (3,)...
            self._block_size = np.prod(s)
        else:
            self._block_size = sum([s.block_size for s in self._array])
        return self._block_size
    @property
    def append_depth(self):
        return self._append_depth
    @append_depth.setter
    def append_depth(self, d):
        inc = d - self._append_depth
        self._append_depth = d
        if not self._is_simple:
            for a in (self._array.values() if isinstance(self._array, OrderedDict) else self._array):
                a.append_depth += inc

    def _invalidate_type_cache(self):
        """Whenever the dtype and stype are changed we can use this to invalidate the cached forms

        :return:
        :rtype:
        """
        self._dtype = None

    def _get_complex_dtype(self):
        # I think all the shape changes get adequately passed down...?
        # Like I think Repeating and friends manage stuff fine
        arr = self._array
        if arr is None:
            # Means we gotta do this recursively
            # The shape should be fed back down to the object's children at this point
            shape = self._stype.shape
            dt = self._stype.dtype
            if isinstance(dt, OrderedDict):
                dt = OrderedDict((k, StructuredType(s) if isinstance(s, type) else s) for k,s in dt.items())
            else:
                dt = tuple(
                    StructuredType(s) if isinstance(s, type) else s for s in self._stype.dtype
                )
            if shape is not None:
                # we have to take our shape and feed it back down to our children
                # the mechanism for this will have to be to take each type in dt and stick our shape onto the
                # front of the dtype's shape
                # one thing about this, is that once we've fed the shape down in
                import copy
                if isinstance(dt, OrderedDict):
                    dt = OrderedDict((k, copy.copy(d)) for k,d in dt.items())
                    for d in dt.values():
                        d.extend_shape(shape)
                else:
                    dt = tuple( copy.copy(d) for d in dt )
                    for d in dt:
                        d.extend_shape(shape)
                st = copy.copy(self._stype)
                st.shape = None
                st.dtype = dt
                self._stype = st
        elif isinstance(arr, OrderedDict):
            dt = OrderedDict((k, a.stype) for k,a in arr.items())
            self._stype = StructuredType(dt, shape=None)
        else:
            dt = tuple(a.stype for a in arr)
            self._stype = StructuredType(dt, shape=None)

        return dt
    @property
    def dtype(self):
        """Returns the core data type held by the StructuredType that represents the array

        :return:
        :rtype:
        """

        if self._dtype is None:
            if self.is_simple:
                self._dtype = self._stype.dtype
            else:
                self._dtype = self._get_complex_dtype()

        return self._dtype

    @property
    def stype(self):
        """Returns the StructuredType that the array holds data for

        :return:
        :rtype:
        """

        if self._dtype is None:
            # allows us to keep the dtype and stype aligned
            fix_me = self.dtype
        return self._stype

    @property
    def array(self):
        if self.is_simple:
            slices = tuple(slice(0, x) for x in self.filled_to if x > 0) # treating 0 as "all"
            try:
                return self._array[slices]
            except IndexError:
                raise ValueError("can't slice array of shape {} to `filled_to` spec {}".format(
                    self._array.shape,
                    self.filled_to
                ))
        else:
            return self._array
    @property
    def _subarrays(self):
        return self._array.values() if isinstance(self._array, OrderedDict) else self._array

    def axis_shape_indeterminate(self, axis):
        """Tries to determine if an axis has had any data placed into it or otherwise been given a determined shape

        :param axis:
        :type axis:
        :return:
        :rtype:
        """
        indet = self.filled_to[axis] == 0
        if indet and self.stype.shape is not None:
            indet = self.stype.shape[axis] is None
        return indet
    @property
    def has_indeterminate_shape(self):
        """Tries to determine if the entire array has a determined shape

        :param axis:
        :type axis:
        :return:
        :rtype:
        """
        if self.is_simple:
            indet = all(f == 0 for f in self.filled_to)
            if indet and self.stype.shape is not None:
                count_None = 0
                for c in self.stype.shape:
                    if c is None:
                        count_None += 1
                        if count_None > 1:
                            break
                indet = count_None == len(self.filled_to) or count_None > 1
            return indet # eh we'll call this enough for now
        else:
            return any( a.has_indeterminate_shape for a in self._subarrays )

    @property
    def filled_to(self):
        if self._is_simple:
            if self._filled_to is None:
                self._filled_to = [ 0 ] * len(self._array.shape)
            return self._filled_to
        else:
            return [ s.filled_to for s in self._subarrays ]
    @filled_to.setter
    def filled_to(self, filling):
        # print(">>>>", self._filled_to)
        if self._is_simple:
            if isinstance(filling, int):
                # print("  >>", 1, filling, self._filled_to)
                self.filled_to[0] = filling
            elif len(filling) < len(self.filled_to):
                # print("  >>", 2, filling, self._filled_to)
                self._filled_to = list(filling) + self._filled_to[len(filling):]
            else:
                # print("  >>", 3, filling, self._filled_to)
                self._filled_to = list(filling)
        else:
            # thread the setting
            raise NotImplementedError("can't apply filling {} to multidim array {} because I'm lazy".format(
                filling,
                self
            ))

        # print(self._filled_to, "<<<<")
    def set_filling(self, amt, axis = 0):
        if self._is_simple:
            if self._filled_to is None:
                _ = self.filled_to # populates it
            self._filled_to[axis] = amt
        else:
            # gotta propagate the filling down to the bottom axis
            for a in self._subarrays:
                a.set_filling(amt, axis = axis)
    def increment_filling(self, inc = 1, axis = 0):
        if self._is_simple:
            if self._filled_to is None:
                _ = self.filled_to # populates it
            self._filled_to[axis] += inc
        else:
            # gotta propagate the filling down to the bottom axis
            for a in self._subarrays:
                a.increment_filling(inc = inc, axis = axis)

    def __len__(self):
        return len(self.array)

    def empty_array(self, shape = None, num_elements = None):
        """Creates empty arrays with (potentially) default elements

        The shape handling rules operate like this:
            if shape is None, we assume we'll initialize this as an array with a single element to be filled out
            if shape is (None,) or (n,) we'll initialize this as an array with multiple elments to be filled out
            otherwise we'll just take the specified shape

        :param num_elements:
        :type num_elements:
        :return:
        :rtype:
        """

        dt = self.dtype
        stype = self.stype
        if shape is None:
            shape = stype.shape if self.shape is None else self.shape
        if num_elements is None:
            num_elements = self._default_num_elements

        if stype.is_simple:
            if shape is None:
                # means we expect to have single object, not a vector of them
                if stype.default is None:
                    arr = np.empty((1, ), dtype=dt)
                else:
                    arr = np.full((1, ), stype.default, dtype=dt)
            else:
                if any(x is None for x in shape):
                    # means we have an array of indeterminate size in that dimension
                    shape = tuple(num_elements if x is None else x for x in shape)
                # might want to add a check to see if dt is valid fo np.array
                if stype.default is None:
                    arr = np.empty(shape, dtype=dt)
                else:
                    arr = np.full(shape, stype.default, dtype=dt)
            return arr
        elif isinstance(dt, OrderedDict):
            return OrderedDict((k, StructuredTypeArray(s)) for k,s in dt.items())
        else:
            return tuple( StructuredTypeArray(s) for s in dt )

    def extend_array(self, axis = None):
        array = self._array
        if isinstance(array, np.ndarray):
            ax = self.extension_axis if axis is None else axis
            empty = self.empty_array(shape = self._array.shape) # should effectively double array size?
            self._array = np.concatenate(
                (
                    array,
                    empty
                ),
                axis=ax
            )
        elif isinstance(array, OrderedDict):
            for arr in array.items():
                arr.extend_array( axis = axis )
        else:
            for arr in array:
                arr.extend_array( axis = axis )

    def __setitem__(self, key, value):
        self.set_part(key, value)
    def set_part(self, key, value):
        """Recursively sets parts of an array if not simple, otherwise just delegates to NumPy

        :param key:
        :type key:
        :param value:
        :type value:
        :return:
        :rtype:
        """
        if self.is_simple:
            # this means that self._array is a plain numpy array

            append_chops = 1 if isinstance(key, int) else len(key) # how many dimensions in we dove for the append

            if isinstance(key, int) and self._array.shape[0] == key:
                self.extend_array(axis=0)
            else:
                for i, k in enumerate(key):
                    if k == self._array.shape[i]:
                        self.extend_array(axis=i)

            if value is not None: # we use None as a placeholder for the default value because we need it for Optional patterns
                if isinstance(value, StructuredTypeArray):
                    value = value.array
                residual_dims = len(self._array.shape) - append_chops

                if isinstance(value, np.ndarray) and residual_dims > 0:
                    if residual_dims == 0:
                        # nothing to do here
                        pass
                    elif residual_dims == 1:

                        value = value.flatten()
                        # we can actually manage to do some padding, so why not do so?
                        if self.axis_shape_indeterminate(append_chops):
                            curr_slices = (slice(None, None), ) * append_chops
                            slices =  curr_slices + ( slice(0, value.shape[0]), )
                            self._array = self._array[slices]
                            # we do
                        else:
                            # we only allow 1D padding for now...
                            num_els = len(value)
                            # try:
                            num_needed = len(self._array[key])
                            # except:
                            #     print("Oh foook", key, self._array.shape)
                            #     raise
                            if num_els < num_needed:
                                if self.padding_mode == 'repeat':
                                    repeats = int(np.ceil(num_needed/num_els))
                                    value = np.tile(value, repeats)[:num_needed]
                                elif self.padding_mode == 'last':
                                    value = np.concatenate((value, np.full(num_needed-num_els, value[-1])))
                                elif self.padding_mode == 'fill':
                                    value = np.concatenate((value, np.full(num_needed-num_els, self.padding_value)))
                                else:
                                    raise StructuredTypeArrayException("unknown padding_mode '{}'".format(
                                        self.padding_mode
                                    ))

                    else:
                        # we can now determine our shape and so we will force the shape
                        take_all = slice(None, None)
                        curr_slices = (take_all, ) * append_chops
                        slices = curr_slices + tuple(
                            slice(0, x) if self.axis_shape_indeterminate(i) else take_all
                            for i, x in enumerate(value.shape)
                        )
                        self._array = self._array[slices]

                    value = value.astype(self.dtype)
                    # string check
                    if self.dtype is str and (self._array.dtype != value.dtype):
                        num_arr = int(str(self._array.dtype).strip("<US|"))
                        num_val = int(str(value.dtype).strip("<US|"))
                        if num_arr < num_val:
                            self._array = self._array.astype(value.dtype)

                    self._array[key] = value
                    fill=self.filled_to

                    chopped_fill = fill[:append_chops]
                    max_chops = list(max(a, s) for a,s in zip(fill[append_chops:], value.shape))
                    self.filled_to = chopped_fill + max_chops
                else:
                    # I think this is what there being no residual dims means
                    # (essentially single-element insert)
                    self._array[key] = value

            if isinstance(key, int) and key == self.filled_to[0]:
                self.filled_to[0] += 1
            elif key == tuple(self.filled_to[:append_chops]):
                self.filled_to[append_chops-1] += 1

        else:
            if isinstance(key, (int, str)):
                key = [ key ] * len(self._array)
            if value is None:
                value = [ None ] * len(self._array)

            if len(value) > len(self._array):
                # means we need to reshape our value arrays since it came from a groups call but groups just returns flat values,
                # not the actual grouping we might want
                shapes = self.shape
                # since we're assigning to a slice I guess this means we're assigning the shape of element 2?
                # so we can just split it by the number of elements we need from shape[1:] and then use np.reshape on that
                # ah but we have pre-populated a number of the np.array rows so really we need to go from shape[2:]
                old_value = value
                value = [ None ]*len(shapes)
                for i, s in enumerate(shapes):
                    sub_shape = s[1:]
                    s_num = np.prod(sub_shape) # get the number of elements
                    if s_num == 1:
                        value[i] = old_value[0]
                        old_value = old_value[1:]
                    else:
                        v = old_value[:s_num]
                        old_value = old_value[s_num:]
                        value[i] = np.reshape(np.array(v, dtype=object), sub_shape)

            for a, k, v in zip(self._array, key, value):
                a[k] = v
    def __getitem__(self, item):
        return self.get_part(item, use_full_array=False)
    def get_part(self, item, use_full_array = True):
        """If simple, delegates to NumPy, otherwise tries to recursively get parts...?
        Unclear how slicing is best handled here.

        :param item:
        :type item:
        :return:
        :rtype:
        """
        if use_full_array:
            arr = self._array
        else:
            arr = self.array
        if self.is_simple:
            return arr[item]
        else:
            if isinstance(item, tuple):
                compound_index = True
                first_thingy = item[0]
            else:
                compound_index = False
                first_thingy = item
            if isinstance(first_thingy, slice) and compound_index:
                bits = arr[first_thingy]
                return [ b[item[1:]] for b in bits ]
            elif isinstance(first_thingy, slice):
                raise NotImplemented("Not sure how I want to slice StructuredType objects yet")
            elif compound_index:
                bit = arr[first_thingy]
                return bit[item[1:]]
            else:
                bit = arr[first_thingy]
                return bit

    def add_axis(self, which = 0, num_elements = None, change_shape = True):
        """Adds an axis to the array, generally used for expanding from singular or 1D data to higher dimensional
        This happens with parse_all and repeated things like that

        :param which:
        :type which:
        :param num_elements:
        :type num_elements:
        :return:
        :rtype:
        """

        # print(">>>>>", self)
        change_shape = True # just gonna see what effect this has...?
        if self.is_simple:
            if self.has_indeterminate_shape:
                import copy
                self._stype = copy.copy(self._stype)

                if self.stype.shape is None:
                    if change_shape:
                        self.stype.shape = (None,)
                        self.filled_to = [0] + self.filled_to
                else:
                    st = self.stype
                    s1 = st.shape
                    st.shape = (None,) + st.shape
                    self._array = None
                    self._array = self.empty_array(num_elements = self._default_num_elements if num_elements is None else num_elements)
                    if not change_shape:
                        self.stype.shape = s1
                    self.filled_to = [0] + self.filled_to

            elif self.stype.shape is not None:
                # if we've already got a flexible shape I think we actually don't need to add an axis...
                # otherwise we add an axis where specified, but in practice which will never not be 0

                shape = self._array.shape # just broadcast the numpy array
                if num_elements is None:
                    num_elements = self._default_num_elements

                new_shape_1 = shape[:which] + (1,) + shape[which:] # we gotta broadcast to the 1 first for some corner cases
                self._array = np.broadcast_to(self._array, new_shape_1)
                new_shape_2 = shape[:which] + (num_elements,) + shape[which:] # now we can fully broadcast
                self._array = np.broadcast_to(self._array, new_shape_2).copy()

                # now we change up the shape of our stype...?
                # maybe this should be possible to have it _not_ do so that we can put all shape
                # changes on the outer-most level...
                if change_shape:
                    import copy
                    # print(self.stype)
                    self._stype = copy.copy(self.stype)
                    if self._stype.shape is None:
                        self._stype.shape = (None,)
                    else:
                        self._stype.shape = self._stype.shape[:which] + (None,) + self._stype.shape[which:]
                    self._dtype = None # resets this flag so the caching is broken

                if self.filled_to[which] > 0:
                    self.filled_to = [1] + self.filled_to # we have one copy of our data already
                else:
                    self.filled_to = [0] + self.filled_to

        else:
            # we'll recursively add axes to the subarrays I think...
            # this might not be entirely in keeping with how a np.ndarray works but I think is acceptable
            arr = self._array
            if isinstance(arr, OrderedDict):
                aiter = arr.values()
            else:
                aiter = arr
            for a in aiter:
                a.add_axis(which, num_elements, change_shape = False)
                # eeeeh is this right?
                # for this to really make sense we'd need to both

            if isinstance(arr, OrderedDict):
                self._stype.dtype = OrderedDict((k, a.stype) for k, a in arr.items())# feels weird to have to correct this but oh well
            else:
                self._stype.dtype = tuple(a.stype for a in arr) # I think this is how the new thing will be typed...

            # now we change up the shape of our stype, hopefully only at the outer level...
            if change_shape:
                import copy
                self._stype = copy.copy(self._stype)
                if self._stype.shape is None:
                    self._stype.shape = (None,)
                else:
                    self._stype.shape = (None,) + self._stype.shape
                self._dtype = None # resets this flag so the caching is broken

        # print(self, "<<<<<")

    def can_cast(self, val):
        """Determines whether val can probably be cast to the right return type and shape without further processing or if that's definitely not possible

        :param val:
        :type val:
        :return:
        :rtype:
        """

        castable = self.is_simple
        if castable:
            try:
                val = np.asarray(val).astype(self.dtype) # make sure we can use .shape
            except ValueError:
                castable = False
            else:
                if not isinstance(val, np.ndarray) or val.shape == ():
                    # can't cast to some shape so we gotta have no shape and be a primitive type
                    castable = self.shape is None
                elif len(self.shape) == 1:
                    # gotta have a vector of values but we can fill without filling the entire thing I'd say
                    castable = len(val.shape) == 1
                else:
                    axis, block_size, remainder, shape = self._get_casting_shape(val)
                    castable = remainder == int(remainder)

        return castable

    def append(self, val, axis=0):
        """
        Puts val in the first empty slot in the array

        :param val:
        :type val:
        :return:
        :rtype:
        """

        axis = axis + max(self.append_depth, 0)
        if self.is_simple:
            pos = tuple(self.filled_to[:axis+1])
        else:
            pos = [tuple(f[:axis+1]) for f in self.filled_to]
        self[pos] = val
        # self._filled_to+=1 # handled automatically by a small bit of cleverness in the filling code

    def _get_casting_shape(self, val, axis = None):
        axis = self.extension_axis if axis is None else axis
        vs = val.shape
        ss = self.shape
        vs_a = vs[:axis] + vs[axis+1:]
        ss_a = ss[:axis] + ss[axis+1:]
        my_stuffs = np.prod(ss_a)
        if vs_a != ss_a and my_stuffs > 0:
            total_stuffs = np.prod(vs)
            remaining_stuff = total_stuffs/my_stuffs
            new_shape = ss[:axis] + (int(remaining_stuff),) + ss[axis+1:]
        else:
            # the my_stuffs == 0 is corner case for super indeterminate shapes...?
            # Or means I need to handle the shape more intelligently...
            # at this point what is there to do? Off of the extension axis none of the stuff has been filled...?
            # does this just mean we basically return the shape of the current thing? This will be used in two places:
            # SA.extend and SA.can_cast
            # in the first case this should basically tell us that we should just call SA.fill?
            # in the latter this should basically be saying "yes, can cast"?
            remaining_stuff = 0
            new_shape = vs
        return axis, my_stuffs, remaining_stuff, new_shape

    def extend(self, val, single = True, prepend = False, axis = None):
        """Adds the sequence val to the array

        :param val:
        :type val:
        :param single: a flag that indicates whether val can be treated as a single object or if it needs to be reshapen when handling in non-simple case
        :type single: bool
        :return:
        :rtype:
        """

        if isinstance(val, StructuredTypeArray):
            val = val.array

        if self.is_simple:
            # we have a minor issue here, where we might not actually have any data in our array and in that case we
            # won't know what 'extend' means
            # in that case I think we would be safe enough delegating to 'fill'
            # print(":::P", self.shape, self._array.shape, val.shape)
            if self.has_indeterminate_shape and len(self.shape) == len(val.shape if isinstance(val, np.ndarray) else val):
                return self.fill(val)


            # print("...?", axis)

            # now check for shape mismatches so they may be corrected _before_ the insertion
            ax, block_size, remainder, new_shape = self._get_casting_shape(val, axis = axis)
            # print(ax)
            malformed = remainder != int(remainder)
            if malformed:
                if int(remainder) == 0:
                    # there is a use case where we fill our array with a smaller, padded form of our data?
                    # we'll do this initially for only like 1D data...?
                    val = val.flatten()
                    num_els = len(val)
                    repeats = int(np.ceil(block_size/num_els))
                    val = np.tile(val, repeats)[:block_size]
                    ax, block_size, remainder, new_shape = self._get_casting_shape(val)
                    malformed = remainder != int(remainder)
                if malformed:
                    raise StructuredTypeArrayException("{}.{}: object with shape '{}' can't be used to extend array of shape '{}' along axis '{}'".format(
                        type(self).__name__,
                        'extend',
                        val.shape,
                        self.shape,
                        ax
                    ))
            val = val.reshape(new_shape).astype(self.dtype)

            filling = self.filled_to[ax] # I should update this so it works for axis != 0 too...
            # print(self._array.shape, val.shape, val)
            self._array = np.concatenate(
                (
                    self._array[:filling],
                    val
                ) if not prepend else (
                    val,
                    self._array[:filling]
                ),
                axis = ax
            )
            self.filled_to[ax] = self._array.shape[ax]
        else:
            if not single and isinstance(val, np.ndarray): # single alone might tell us we have an issue...
                if len(val[0]) == len(self._array):
                    # just need to transpose the groups, basically,
                    gg = val.T
                else:
                    blocks = [b.block_size for b in self._subarrays]
                    # we'll assume val is an np.ndarray for now since that's the most common case
                    # but this might not work in general...
                    gg = [ None ] * len(blocks)
                    sliced = 0
                    for i, b in enumerate(blocks):
                        if b == 1:
                            gg[i] = val[:, sliced]
                        else:
                            gg[i] = val[:, sliced:sliced+b]
                        sliced += b
                val = gg
            for a, v in zip(self._array, val):
                a.extend(v, axis = axis)

    # def fill_to(self, level, fill_value = None):
    #     """Sets the filled_to parameter level of the array to level, optionally filling with a value
    #
    #     :param level:
    #     :type level:
    #     :param fill_value:
    #     :type fill_value:
    #     :return:
    #     :rtype:
    #     """
    #
    #     base_level = self.filled_to[0]
    #     if level > base_level:
    #         if self.is_simple:
    #             self.filled_to[0] = level
    #         else:
    #             for a in self._subarrays:
    #                 a.fill_to(level, fill_value=fill_value)
    #     # uh... what do about the fill_value?
    #     if fill_value is not None:
    #         raise NotImplementedError("I still haven't implemented fill_value yet ;_;")

    def fill(self, array):
        """Sets the result array to be the passed array

        :param array:
        :type array: str | np.ndarray
        :return:
        :rtype:
        """

        # I do want to make it so that I don't lose my shape, though... or at least I don't lose the
        # entire structure of my shape. I can't try to preserve _too_ many axes, but I can at least preserve the number
        # of them I think...

        if isinstance(array, str):
            array = self.cast_to_array(array)
        elif isinstance(array, StructuredTypeArray):
            array = array.array

        if self._is_simple:
            # not sure why the array _wouldn't_ be an np.ndarray but there's a lot going on and I'm tired and don't
            # want to figure it out
            if isinstance(array, np.ndarray):
                # one big thing we have to watch out for is shrinking the array along some _determined_ axis
                # it doesn't matter what we do to indeterminate axes, but determined ones matter
                if self.dtype is not None:
                    array = array.astype(self.dtype)
                shp = self._stype.shape
                if shp is None or all(a is None for a in shp):
                    self._array = array
                else:
                    ashp = array.shape
                    if len(shp) > len(ashp):
                        # we'll force it to have the right number of axes?
                        # to be honest this is subtle and I'm not going to handle it all properly right now
                        # definitely a source of bugs in the future
                        # I would also probably need to handle the higher-dimensional case too

                        new_shp = [ 1 ] * len(shp)
                        indet_axis_count = len(shp) - len(ashp)
                        which_array_shp = 0
                        for i in range(len(shp)):
                            # we will either pull from shp if the thing is a determined axis
                            # or from ashp if not, but only after we've exhausted the index_axis_count
                            if shp[i] is not None:
                                new_shp[i] = shp[i]
                            elif indet_axis_count > 0:
                                indet_axis_count -= 1
                            else:
                                new_shp[i] = ashp[which_array_shp]
                                which_array_shp += 1

                        array = array.reshape(new_shp)
                        self._array = array
                    elif len(shp) == len(ashp):
                        # we should _probably_ check to make sure we're not breaking things but... meh
                        self._array = array
                    else:
                        raise NotImplementedError("Filling data of shape {} into {}, but I don't know how to coerce a higher dimensional shape into a lower dimensional one".format(
                            shp,
                            ashp
                        ))

                    # should we also force the values to be right...?

            else:
                # print(">>>>", array)
                self._array = np.array(array, dtype=str)

            # at this point we should deal with the filling level that we already had, I think...
            ft = self.filled_to[0]
            if ft > 0:
                # this isn't going to respect the axis quite right but I
                # don't want to deal with that yet...
                # basically just gonna add stuff to make it so we have the right
                # shape for the claimed filling level
                self._array = np.concatenate(
                    (
                        self._array[:ft],
                        self._array
                    )
                )
            self.filled_to = self._array.shape
        else:
            for arr, data in zip(self._array, array):
                arr.fill(data)
            self._filled_to = None

    def cast_to_array(self, txt):
        """Casts a string of things with a given data type to an array of that type and does some optional
        shape coercion

        :param txt:
        :type txt: str | iterable[str]
        :return:
        :rtype:
        """
        if self.is_simple:
            if len(txt.strip()) == 0:
                arr = np.array([], dtype=self._stype)
            else:
                try:
                    # we'll try the base conversion first just assuming we got a number or whatever
                    # and it managed to filter through the code to here
                    arr = np.array([txt], dtype=self._stype)
                except TypeError:
                    import io
                    arr = np.loadtxt(io.StringIO(txt), dtype=self._stype)
                    shape = np.array(self.shape)
                    axis = self.extension_axis
                    if shape is not None and shape[axis] > 0: # make sure arr needs to be reshaped...
                        arr = arr.flatten()
                        num_els = len(arr)
                        # the number of elements that the length of the parsed out array must be divisible by
                        num_not_along_axis = np.prod(shape) / shape[axis]
                        if num_els % num_not_along_axis == 0:
                            # means we can cleanly reshape it once we know the target shape
                            shape[self.extension_axis] = num_els / num_not_along_axis
                            arr = arr.reshape(shape)
                        # should we raise an error if not?

        else: #we'll assume we got some block of strings since there's no reason to put a parser here...
            arr = [a.cast_to_array(t) for a, t in zip(self._array, txt)]

        return arr

    def __repr__(self):
        return "{}(shape={}, dtype={})".format(
            type(self).__name__,
            self.shape,
            self.dtype
        )


class StructuredTypeArrayException(Exception):
    pass